# UnityBase Benchmarking Test

Benchmarks for UnityBase for a scenarios from the [TechEmpower Web Framework Benchmarks](https://www.techempower.com/benchmarks/).

# Running the benchmarks

The benchmark repo is set up to work against the latest UnityBase version so make sure you read through the
following details to help you get started.

## The scenarios
Following are the details of each of the scenarios the server application contains implementations for and thus can be benchmarked:

| url | Name | Description |
| :--- | :--- | :--- |
| /plaintext | Plaintext | This test is an exercise of the request-routing fundamentals only, designed to demonstrate the capacity of high-performance platforms in particular. Requests will be sent using HTTP pipelining. |
| /json | JSON | This test exercises the framework fundamentals including keep-alive support, request routing, request header parsing, object instantiation, JSON serialization, response header generation, and request count throughput. |
| /dbRaw | Single Query Raw | This test exercises the framework's random number generator, database driver, and database connection pool. |
| /db | Single Query ORM | As for the single query raw test above but using UBQL as the ORM. |
| /queriesRaw?queries=20 | Multiple Queries Raw | Multiple rows are fetched to more dramatically punish the database driver and connection pool. |
| /queries?queries=20 | Multiple Queries ORM | As for the multiple query raw test above but using UBQL as the ORM. |
| /fortunesRaw | Fortunes Raw | This test exercises the database connectivity, dynamic-size collections, sorting, server-side templates, XSS countermeasures, and character encoding. |
| /fortunes | Fortunes ORM | As for the Fortunes raw test above but using UBQL as the ORM. |

## Setting up the UnityBase
Download and install the latest UnityBase Standard Edition setup from https://unitybase.info/download-files.html (15Mb)
for your Linux distribution. Benchmarking on Windows id *deprecated*.

## Verify installation
```shell
ub -e "console.log('It is work')"
```
This command evaluate a JavaScript and output to console `It is work`.


## Setting up the application

Pre-requirements: [install `nodejs` and `npm`](https://nodejs.org/en/download/package-manager/) for your platform. Version 16+ is required.

UnityBase itself *DO NOT* require nodejs to be installed, we need it just to install application dependencies.
For production environment recommended method to deploy app without installing nodejs on production server is
to use `ub-pack/ub-deploy` ad described in [Manage production environments](https://unitybase.info/api/server-v5/tutorial-production_env.html) 

Clone this repo and execute
```shell
cd apps/benchmark
npm i
./prepare.sh
```
This will:
- install necessary JS packages
- create an empty SQLite3 database
- create a tables for a application domain
- fill tables with data necessary for tests


## Generating Load
It's best to generate load from a completely separate machine from the server if you can, to avoid resource contention during the test.

Techempower use the [wrk](https://github.com/wg/wrk) load generation tool to generate the load for our benchmark runs. 
It's the best tool we've found for the job and supports HTTP pipelining (used by the plaintext scenario) via its scripting interface. 

If you have a Golang, [Go hey](https://github.com/rakyll/hey) can be used instead of wrk

The most close to real life test case is `fortunes`. Load for fortunes can be generated using

Start UnityBase using:
```shell
cd apps/benchmark
ub
```

Run locally:
```shell
wrk -H 'Host: 10.0.0.1' -H 'Connection: keep-alive' \
  -H 'Accept: application/json,text/html;q=0.9,application/xhtml+xml;q=0.9,application/xml;q=0.8,*/*;q=0.7' \
  --latency -d 15 -c 512 --timeout 8 -t 28 "http://localhost:8881/fortunes"
```

## Results
For AMD Ryzen 5 5600G fortunes result is ~15k RPS:

```shell
wrk -H 'Host: 10.0.0.1' -H 'Accept: application/json,text/html;q=0.9,application/xhtml+xml;q=0.9,application/xml;q=0.8,*/*;q=0.7' -H 'Connection: keep-alive' --latency -d 15 -c 512 --timeout 8 -t 28 "http://localhost:8881/fortunes"
Running 15s test @ http://localhost:8881/fortunes
  28 threads and 512 connections
  Thread Stats   Avg      Stdev     Max   +/- Stdev
    Latency    32.72ms    3.18ms  80.32ms   72.16%
    Req/Sec   550.13     30.60   670.00     93.56%
  Latency Distribution
     50%   32.85ms
     75%   34.75ms
     90%   36.44ms
     99%   39.25ms
  230709 requests in 15.08s, 306.05MB read
Requests/sec:  15302.10
Transfer/sec:     20.30MB
```

On real productions of our Document Management System we discover what 1000 logged-in users typically generates 100 RPS,
so 15K RPS is enough for 150 000 users.

## More than benchmark
In fact, we got not just a test, but the complete application. 
During of `prepare.sh` execution UnityBase take a config `ubConfig.json` [JSON schema of config](https://unitybase.info/models/UB/docson/index.html#../schemas/ubConfig.schema.json)
and:

 - create a empty database using [cmd/initDB command](https://unitybase.info/api/serverNew/module-cmd_initDB.html)
 - analyse the [application domain](https://unitybase.info/docs/design.html#Models) and generate a database tables using [cmd/generateDDL](https://unitybase.info/api/serverNew/module-cmd_generateDDL.html) command
 - fill the initial data for each model, included in the domain using [cmd/initialize](https://unitybase.info/api/serverNew/module-cmd_initialize.html) command

### User interface
There is a `adminUI` - a Single Page Rich Internet Application.

![adminUI interface sample](https://github.com/UnityBaseJS/benchmarks/raw/master/img/adminUI.png)

If you already execute a `prapare.cmd` - start a UB in a developer mode with logging enabled by typing

    ub -dev -cfg ubConfigWLog.json

This command will start a UnityBase Application server in a singe thread GUI mode with JavaScript debugger & logging.
Select `Service -> Run in browser` from a menu and explore a `adminUI` application (Google Chrome required). Use username `admin` and password `admin`.

Everything you will see in the `adminUI` is either generated based on the entity definitions, or customizable directly from the `adminUI`.

### Debugging
You can debug a server-side JavaScript in the GUI - select a server thread from the ThreadID combo, in the `Source` tab search for the `methods.js`.
Double click to see the sources. Set a breakpoint inside the `function db`, run a db method from a browser `http://localhost:888/db` and debug a method.

### Security
For a testing purpose all security are disabled for a techempower endpoints, but all other endpoint are secure - see [this page for details](https://unitybase.info/api/serverNew/tutorial-security.html)

### Database support
For a simplicity techempower tests executed using build-in SQLite3 database. But you can use PostgreSQL, Oracle or MS SQL.
Modify a connections section of a config (ubConfig**YourDatabase**.json) file for database you need and ron (sample for a Postgres):

```shell
UB_CFG=ubConfigPostgre.json
DBA=postgres
DBA_PWD=postgres
prepare.sh
ub -cfg ubConfigPostgre.json
```

See a description of config in the [JSON schema of config].


[JSON schema of config]: https://unitybase.info/models/UB/docson/index.html#../schemas/ubConfig.schema.json