const _ = require('lodash')
const ElementUI = require('element-ui')
const UB = require('@unitybase/ub-pub')

const SHOW_TIMEOUT = 500
const HIDE_TIMEOUT = 500

class LoadingOverlayManager {
  constructor () {
    this._loaderServiceInstance = undefined
    this._showTimeout = undefined
    this._hideTimeout = undefined

    /**
     * @type {Object<string, Array>}
     * @private
     */
    this._loadersByText = {}
  }

  /**
   * Add a loader from loaders list
   *
   * @param {string} text
   */
  show (text) {
    const exists = this._loadersByText[text]
    if (!exists) {
      this._loadersByText[text] = []
    }

    this._loadersByText[text].push({
      text,
      started: Date.now()
    })

    this._calcText()
  }

  /**
   * Remove a loader from loaders list
   *
   * @param {string} text
   */
  hide (text) {
    const exists = this._loadersByText[text]
    if (!exists) {
      console.warn('There is no loader with text %s', text)
      return
    }
    exists.shift()
    if (exists.length === 0) {
      delete this._loadersByText[text]
    }

    this._calcText()
  }

  /**
   * Recalculate text to show as a loading text after add or remove a loader
   *
   * @private
   */
  _calcText () {
    const text = _.maxBy(
      Object.entries(this._loadersByText),
      ([, loaders]) => _.maxBy(loaders, 'started')
    )?.[0]

    if (this._text !== text) {
      this._text = text

      const needShow = !!text

      if (needShow) {
        if (this._loaderServiceInstance) {
          this._loaderServiceInstance.text = this._buildText(text)
          if (this._hideTimeout) {
            this._removeHideTimeout()
          }
        } else if (!this._showTimeout) {
          this._setShowTimeout()
        }
      } else {
        if (!this._loaderServiceInstance && this._showTimeout) {
          this._removeShowTimeout()
        }
        if (this._loaderServiceInstance && !this._hideTimeout) {
          this._setHideTimeout()
        }
      }
    }
  }

  _setShowTimeout () {
    this._showTimeout = setTimeout(
      () => {
        this._showTimeout = undefined
        this._showLoader()
      },
      SHOW_TIMEOUT
    )
  }

  _removeShowTimeout () {
    clearTimeout(this._showTimeout)
    this._showTimeout = undefined
  }

  _setHideTimeout () {
    this._hideTimeout = setTimeout(
      () => {
        this._hideTimeout = undefined
        this._hideLoader()
      },
      HIDE_TIMEOUT
    )
  }

  _removeHideTimeout () {
    clearTimeout(this._hideTimeout)
    this._hideTimeout = undefined
  }

  _showLoader () {
    if (this._loaderServiceInstance) {
      this._loaderServiceInstance.close()
    }
    this._loaderServiceInstance = ElementUI.Loading.service({ text: this._buildText(this._text) })
  }

  _hideLoader () {
    if (this._loaderServiceInstance) {
      this._loaderServiceInstance.close()
      this._loaderServiceInstance = undefined
      this._text = undefined
    }
  }

  _buildText (textResourceId) {
    return UB.i18n(textResourceId)
  }
}

/**
 * Subscribe to 'portal:loader' event and manage the loader overlay using ElementUI
 */
function subscribe () {
  const manager = new LoadingOverlayManager()

  window.$App.on('portal:loader', ({ show, text = 'portal.loader.default' }) => {
    if (show) {
      manager.show(text)
    } else {
      manager.hide(text)
    }
  })
}

module.exports = {
  subscribe
}
