/**
 * Restrictions for role `LocalOrgManager` on modify entities
 * `uba-user|uba_userrole|uba_usergroup` and `org_unit`
 */

const UB = require('@unitybase/ub')
const Session = UB.Session
const UBA_COMMON = require('@unitybase/base').uba_common

global.uba_user.on('update:before', ctx => {
  if (ctx.externalCall) checkPermissionForUser(ctx.mParams.execParams.ID)
})
global.uba_user.on('delete:before', ctx => {
  if (ctx.externalCall) checkPermissionForUser(ctx.mParams.execParams.ID)
})

global.uba_userrole.on('insert:before', checkPermissionForUserDetails)
global.uba_userrole.on('update:before', checkPermissionForUserDetails)
global.uba_userrole.on('delete:before', checkPermissionForUserDetails)

global.uba_usergroup.on('insert:before', checkPermissionForUserDetails)
global.uba_usergroup.on('update:before', checkPermissionForUserDetails)
global.uba_usergroup.on('delete:before', checkPermissionForUserDetails)

global.org_unit.on('update:before', checkPermissionForOrgUnit)
global.org_unit.on('insert:before', checkPermissionForOrgUnit)
global.org_unit.on('delete:before', checkPermissionForOrgUnit)

/**
 * Load userID from detailed entity (uba_userrole|uba_usergroup) and call `checkPermissionForUser`
 * @param {ubMethodParams} ctx
 */
function checkPermissionForUserDetails (ctx) {
  if (!ctx.externalCall || !Session.hasRole('LocalOrgManager') || UBA_COMMON.isSuperUser())  return
  const execParams = ctx.mParams.execParams
  const detailEntity = ctx.dataStore.entityCode || ctx.mParams.entity // TODO - remove hack after adding ctx.entityCode
  const userID = execParams.userID || UB.Repository(detailEntity)
    .attrs('userID')
    .where('ID', '=', execParams.ID)
    .selectScalar() || null
  if (userID) checkPermissionForUser(userID)
}

/**
 * User with the "Local Administrator" role can only modify records of subordinates in own organization
 * @param {Number} userID
 * @return {boolean}
 */
function checkPermissionForUser (userID) {
  if (!Session.hasRole('LocalOrgManager') || UBA_COMMON.isSuperUser()) return true
  const eosTreePaths = UB.Repository('org_employeeonstaff')
    .attrs('staffUnitID.parentID.mi_treePath')
    .where('employeeID.userID', '=', userID)
    .selectAsObject({
      'staffUnitID.parentID.mi_treePath': 'treePath'
    })
  const isSubordinated = eosTreePaths.some(item => item.treePath.startsWith(Session.uData.ownOrganization.treePath))
  if (!isSubordinated) {
    throw new UB.UBAbort('<<<LocalOrgManager.userUpdateDeny>>>')
  }
}


/**
 * User with the "LocalOrgManager" role can only modify records of subordinates in own organization
 * @param {ubMethodParams} ctx
 * @return {boolean}
 */
function checkPermissionForOrgUnit (ctx) {
  if (!Session.hasRole('LocalOrgManager') || UBA_COMMON.isSuperUser()) return true
  const mParams = ctx.mParams
  const execParams = mParams.execParams

  if (mParams.method === 'insert' && !execParams.parentID) throw new UB.UBAbort('AccessDeny')

  let currentTreePath
  let newTreePath
  if (mParams.method !== 'insert') {
    currentTreePath = UB.Repository('org_unit')
      .attrs('mi_treePath')
      .selectById(execParams.ID).mi_treePath
  }
  if (execParams.parentID) {
    newTreePath = UB.Repository('org_unit')
      .attrs('mi_treePath')
      .selectById(execParams.parentID).mi_treePath
  }

  if ((currentTreePath && !currentTreePath.startsWith(Session.uData.ownOrganization.treePath)) ||
    (newTreePath && !newTreePath.startsWith(Session.uData.ownOrganization.treePath))) {
    throw new UB.UBAbort('<<<LocalOrgManager.orgUnitUpdateDeny>>>')
  }
}
